﻿// Copyright 2011-2020 ReflexSystems

using System;
using System.ComponentModel;
using System.Diagnostics;
using System.Globalization;
using System.Runtime.Serialization;

namespace StimulsoftValueType.ValueType
{
    /// <summary>
    /// An object representing an amount in a currency
    /// </summary>
    [Serializable]
    [DebuggerDisplay("{" + nameof(DebuggerDisplay) + ", nq}")]
    [TypeConverter(typeof(MoneyConverter))]
    public readonly struct Money : IEquatable<Money>, IComparable, IComparable<Money>, ISerializable, IFormattable, IConvertible, IHaveCurrencyCode, IDeserializationCallback
    {
        /// <summary>
        /// The currency code to use when the amount is in an unspecified currency
        /// </summary>
        public const string UNSPECIFIED_CURRENCY_CODE = "XXX";

        /// <summary>
        /// Creates a new <see cref="Money"/> object using the specified amount and currency code
        /// </summary>
        /// <param name="amount">The amount</param>
        /// <param name="currencyCode">The currency code to use. Can be null and will default to 'unspecified'</param>
        public Money(decimal amount, string currencyCode)
        {
            currencyCode ??= UNSPECIFIED_CURRENCY_CODE;
            
            if (currencyCode.Length != 3)
                throw new ArgumentException("The currency code should be exactly 3 characters.", nameof(currencyCode));

            this.Amount = amount;
            this.CurrencyCode = currencyCode;
        }

        /// <summary>
        /// Creates a new <see cref="Money"/> object using the specified amount and object that implements the <see cref="IHaveCurrencyCode"/> interface
        /// </summary>
        /// <param name="amount">The amount</param>
        /// <param name="haveCurrencyCode">The object implementing <see cref="IHaveCurrencyCode"/></param>
        public Money(decimal amount, IHaveCurrencyCode haveCurrencyCode) : this(amount, haveCurrencyCode.CurrencyCode)
        {
        }

        /// <summary>
        /// Creates a new <see cref="Money"/> object using the specified amount and copies the currency code from the supplied <see cref="Money"/> object
        /// </summary>
        /// <param name="amount">The amount</param>
        /// <param name="money">The <see cref="Money"/> object</param>
        public Money(decimal amount, in Money money) : this(amount, money.CurrencyCode)
        {
        }

        /// <summary>
        /// This constructor will be used when a <see cref="Money"/> is deserialized.
        /// </summary>
        public Money(SerializationInfo info, StreamingContext context)
        {
            this.Amount = (decimal)info.GetValue(MoneyConstants.AMOUNT_COLUMN, typeof(decimal));
            this.CurrencyCode = (string)info.GetValue(MoneyConstants.CURRENCY_CODE_COLUMN, typeof(string));
        }

        /// <summary>
        /// This method will be called when a <see cref="Money"/> is serialized.
        /// </summary>
        public void GetObjectData(SerializationInfo info, StreamingContext context)
        {
            info.AddValue(MoneyConstants.AMOUNT_COLUMN, this.Amount, typeof(decimal));
            info.AddValue(MoneyConstants.CURRENCY_CODE_COLUMN, this.CurrencyCode ?? UNSPECIFIED_CURRENCY_CODE, typeof(string));
        }

        /// <summary>
        /// The amount
        /// </summary>
        public decimal Amount { get; }
        
        /// <summary>
        /// The currency code
        /// </summary>
        public string CurrencyCode { get; }

        /// <inheritdoc />
        public bool Equals(Money other)
        {
            if (other.CurrencyCode is null)
                return this.CurrencyCode is null;

            if (this.CurrencyCode is null)
                return false;

            return this.Amount == other.Amount && this.CurrencyCode.Equals(other.CurrencyCode);
        }

        /// <inheritdoc />
        public override bool Equals(object obj)
        {
            if (obj is null)
                return false;

            return obj is Money money && this.Equals(money);
        }

        /// <inheritdoc />
        public override int GetHashCode()
        {
            unchecked
            {
                var hash = 17;
                hash = hash * 31 + this.Amount.GetHashCode();
                hash = hash * 31 + this.CurrencyCode.GetHashCode();
                return hash;
            }
        }

        /// <inheritdoc />
        public int CompareTo(object obj) =>
            obj switch
            {
                null => 1,
                int value => value.CompareTo((int)this.Amount), // mainly used to make comparisons in the report engine easier
                double value => value.CompareTo((double)this.Amount), // mainly used to make comparisons in the report engine easier
                decimal value => value.CompareTo(this.Amount), // mainly used to make comparisons in the report engine easier
                Money money => this.CompareTo(money),
                _ => throw new ArgumentException("Object is not a Money object or one of the supported types (int, double, decimal)")
            };

        /// <inheritdoc />
        public int CompareTo(Money other)
        {
            if (this < other)
                return -1;
            return this > other ? 1 : 0;
        }

        /// <summary>
        /// Returns the amount formatted as a number in the current <see cref="NumberFormatInfo"/>
        /// </summary>
        public override string ToString() => FormatMoney(this);

        /// <summary>
        /// Returns the amount formatted as specified in the format string and using the current <see cref="NumberFormatInfo"/>
        /// </summary>
        public string ToString(string format) => FormatMoney(this, format);

        /// <summary>
        /// Will determine if this instance is in an unspecified currency.
        /// </summary>
        /// <returns>true if the instance is in an unspecified currency; false otherwise</returns>
        public bool IsInUnspecifiedCurrency => this.CurrencyCode is null || this.CurrencyCode == UNSPECIFIED_CURRENCY_CODE;

        /// <summary>
        /// Returns the amount as a double
        /// </summary>
        [Obsolete("This is for the report engine. Use Money.Amount instead", false)]
        public static implicit operator double(Money money) => decimal.ToDouble(money.Amount);

        /// <summary>
        /// Returns the amount
        /// </summary>
        [Obsolete("This is for the report engine. Use Money.Amount instead", false)]
        public static implicit operator decimal(Money money) => money.Amount;

        /// <summary>
        /// Creates a new <see cref="Money"/> instance with the specified currency code and the amount set to zero.
        /// </summary>
        /// <param name="currencyCode">An object implementing <see cref="IHaveCurrencyCode"/></param>
        public static Money Zero(IHaveCurrencyCode currencyCode) => new Money(0, currencyCode.CurrencyCode);

        /// <summary>
        /// Creates a new <see cref="Money"/> instance with the specified currency code and the amount set to zero.
        /// </summary>
        /// <param name="currencyCode">A string that represents a 3 letter currency code</param>
        public static Money Zero(string currencyCode) => new Money(0, currencyCode);

        /// <summary>
        /// Creates a new <see cref="Money"/> instance with unspecified currency code and the amount set to the specified amount.
        /// </summary>
        /// <param name="amount">The amount</param>
        public static Money Unspecified(decimal amount) => new Money(amount, UNSPECIFIED_CURRENCY_CODE);

        /// <summary>
        /// Creates a new <see cref="Money"/> instance with unspecified currency code and the amount set to zero.
        /// </summary>
        public static Money Unspecified() => Zero(UNSPECIFIED_CURRENCY_CODE);

        public static bool operator ==(in Money left, in Money right) => Equals(left, right);
        public static bool operator ==(in Money left, in decimal right) => left.Amount == right;
        public static bool operator ==(in Money? left, in decimal right) => left?.Amount == right;

        public static bool operator !=(in Money left, in Money right) => !Equals(left, right);
        public static bool operator !=(in Money left, in decimal right) => left.Amount != right;
        public static bool operator !=(in Money? left, in decimal right) => left?.Amount != right;

        public static Money operator %(in Money left, in Money right)
        {
            var currencyTemplate = GetCurrencyTemplate(left, right);
            return new Money(left.Amount % right.Amount, currencyTemplate);
        }

        public static Money operator %(in Money left, in decimal right) =>
            new Money(left.Amount % right, left.CurrencyCode);

        public static bool operator >(in Money left, in Money right)
        {
            AssertSameCurrency(left, right);
            return left.Amount > right.Amount;
        }

        public static bool operator >(in Money left, in decimal right) =>
            left.Amount > right;

        public static bool operator >=(in Money left, in Money right)
        {
            AssertSameCurrency(left, right);
            return left.Amount >= right.Amount;
        }

        public static bool operator >=(in Money left, in decimal right) =>
            left.Amount >= right;

        public static bool operator <(in Money left, in Money right)
        {
            AssertSameCurrency(left, right);
            return left.Amount < right.Amount;
        }

        public static bool operator <(in Money left, in decimal right) =>
            left.Amount < right;

        public static bool operator <=(in Money left, in Money right)
        {
            AssertSameCurrency(left, right);
            return left.Amount <= right.Amount;
        }

        public static bool operator <=(in Money left, in decimal right) =>
            left.Amount <= right;

        public static Money operator +(in Money left, in Money right)
        {
            var currencyTemplate = GetCurrencyTemplate(left, right);
            return new Money(left.Amount + right.Amount, currencyTemplate);
        }

        public static Money operator +(in Money left, in decimal right) =>
            new Money(left.Amount + right, left.CurrencyCode);

        public static Money operator +(in decimal left, in Money right) =>
            new Money(left + right.Amount, right.CurrencyCode);

        public static Money operator -(in Money left, in Money right)
        {
            var currencyTemplate = GetCurrencyTemplate(left, right);
            return new Money(left.Amount - right.Amount, currencyTemplate);
        }

        public static Money operator -(in Money left, in decimal right) =>
            new Money(left.Amount - right, left.CurrencyCode);

        public static Money operator -(in decimal left, in Money right) =>
            new Money(left - right.Amount, right.CurrencyCode);

        public static Money operator *(in Money left, in decimal right) =>
            new Money(left.Amount * right, left.CurrencyCode);

        public static Money operator *(in decimal left, in Money right) =>
            new Money(left * right.Amount, right.CurrencyCode);

        public static Money operator *(in Money left, in Money right)
        {
            var currencyTemplate = GetCurrencyTemplate(left, right);
            return new Money(left.Amount * right.Amount, currencyTemplate);
        }

        public static Money operator /(in Money left, in Money right)
        {
            var currencyTemplate = GetCurrencyTemplate(left, right);
            return new Money(left.Amount / right.Amount, currencyTemplate);
        }

        public static Money operator /(in Money left, in decimal right) =>
            new Money(left.Amount / right, left.CurrencyCode);

        public static Money operator /(in decimal left, in Money right) =>
            new Money(left / right.Amount, right.CurrencyCode);

        public static Money operator -(in Money left) =>
            new Money(-left.Amount, left.CurrencyCode);
        
        public static Money operator --(in Money left) =>
            new Money(left.Amount - 1, left.CurrencyCode);

        public static Money operator ++(in Money left) =>
            new Money(left.Amount + 1, left.CurrencyCode);

        /// <summary>
        /// Returns the <see cref="Money"/> object with the most specific currency code to be used as a template to determine the currency code
        /// This allows to do operations on a <see cref="Money"/> object where either the left or the right side is in unspecified currency
        /// It still won't allow operations where the left and right hand side have different currency codes
        /// </summary>
        private static Money GetCurrencyTemplate(in Money left, in Money right)
        {
            if (left.IsInUnspecifiedCurrency || right.IsInUnspecifiedCurrency)
            {
                var currencyTemplate = left.IsInUnspecifiedCurrency ? right : left;
                return currencyTemplate.IsInUnspecifiedCurrency ? Unspecified() : currencyTemplate;
            }

            AssertSameCurrency(left, right);

            return left;
        }

        /// <summary>
        /// Compares the currency codes of two money instances and will throw when they don't match.
        /// </summary>
        /// <param name="first">The first <see cref="Money"/> object</param>
        /// <param name="second">The second <see cref="Money"/> object</param>
        public static void AssertSameCurrency(in Money first, in Money second)
        {
            if (first.CurrencyCode != second.CurrencyCode)
                throw new ArgumentException($"Currency code not equal. {first.CurrencyCode} != {second.CurrencyCode}");
        }

        void IDeserializationCallback.OnDeserialization(object sender)
        {
            if (this.CurrencyCode is null)
                throw new SerializationException("CurrencyCode is required");
        }

        TypeCode IConvertible.GetTypeCode() => TypeCode.Object;

        bool IConvertible.ToBoolean(IFormatProvider provider) => throw new InvalidCastException();

        byte IConvertible.ToByte(IFormatProvider provider) => throw new InvalidCastException();

        char IConvertible.ToChar(IFormatProvider provider) => throw new InvalidCastException();

        DateTime IConvertible.ToDateTime(IFormatProvider provider) => throw new InvalidCastException();

        decimal IConvertible.ToDecimal(IFormatProvider provider) => this.Amount;

        double IConvertible.ToDouble(IFormatProvider provider) => (double)this.Amount;

        short IConvertible.ToInt16(IFormatProvider provider) => throw new InvalidCastException();

        int IConvertible.ToInt32(IFormatProvider provider) => throw new InvalidCastException();

        long IConvertible.ToInt64(IFormatProvider provider) => throw new InvalidCastException();

        sbyte IConvertible.ToSByte(IFormatProvider provider) => throw new InvalidCastException();

        float IConvertible.ToSingle(IFormatProvider provider) => throw new InvalidCastException();

        object IConvertible.ToType(Type conversionType, IFormatProvider provider) => throw new InvalidCastException();

        ushort IConvertible.ToUInt16(IFormatProvider provider) => throw new InvalidCastException();

        uint IConvertible.ToUInt32(IFormatProvider provider) => throw new InvalidCastException();

        ulong IConvertible.ToUInt64(IFormatProvider provider) => throw new InvalidCastException();
        
        private string DebuggerDisplay => this.CurrencyCode is null ? "*** Invalid ***" : FormatMoney(this, "C");
        
        /// <summary>
        /// Formats this instance using the specified <see cref="IFormatProvider"/>
        /// </summary>
        public string ToString(IFormatProvider formatProvider) => FormatMoney(this, null, formatProvider);

        /// <inheritdoc />
        public string ToString(string format, IFormatProvider formatProvider) => FormatMoney(this, format, formatProvider);

        private static string FormatMoney(in Money money, string format = null, IFormatProvider formatProvider = null)
        {
            if (format is null || !format.StartsWith("C"))
                return money.Amount.ToString(format, formatProvider);

            formatProvider ??= NumberFormatInfo.CurrentInfo;

            var numberFormatInfo = formatProvider switch
            {
                CultureInfo cultureInfo => cultureInfo.NumberFormat,
                NumberFormatInfo formatInfo => formatInfo,
                _ => throw new InvalidOperationException()
            };

            if (format == "C")
                format += numberFormatInfo.CurrencyDecimalDigits;

            if (!int.TryParse(format.Substring(1), out var decimalDigits))
                throw new FormatException($"The format {format} cannot be parsed");

            return Math.Round(money.Amount, decimalDigits, MidpointRounding.AwayFromZero).ToString("N" + decimalDigits, formatProvider) + " " + money.CurrencyCode;
        }

        /// <summary>
        /// Performs rounding on the specified <see cref="Money"/> object.
        /// </summary>
        /// <param name="money">The <see cref="Money"/> object to perform rounding on</param>
        /// <param name="decimals">The number of decimals</param>
        /// <param name="mode">The rounding mode</param>
        /// <returns>A new <see cref="Money"/> object where the amount is rounded.</returns>
        public static Money Round(in Money money, in int decimals, MidpointRounding mode)
        {
            var amount = decimal.Round(money.Amount, decimals, mode);
            return new Money(amount, money.CurrencyCode);
        }

        /// <summary>
        /// Returns the absolute value of the money object.
        /// </summary>
        /// <param name="money">The <see cref="Money"/> object</param>
        /// <returns>A new <see cref="Money"/> object where the amount has an absolute value</returns>
        public static Money Abs(in Money money) =>
            new Money(Math.Abs(money.Amount), money.CurrencyCode);

        /// <summary>
        /// Takes the decimal value as amount when it's available or returns the value of the specified <see cref="Money"/> object as a new <see cref="Money"/> object
        /// </summary>
        /// <param name="left">The decimal value which can be null</param>
        /// <param name="right">The <see cref="Money"/> object</param>
        /// <returns>A new <see cref="Money"/> with the currency code taken from the specified <see cref="Money"/> object and the amount taken from either the decimal or the specified <see cref="Money"/> object</returns>
        public static Money Coalesce(in decimal? left, in Money right) =>
            new Money(left ?? right.Amount, right.CurrencyCode);

        /// <summary>
        /// Parses the value and attempts to create a new <see cref="Money"/> from it.
        /// </summary>
        /// <param name="s">The value to parse</param>
        /// <param name="result">The returned <see cref="Money"/> object. Will be in unspecified state when the value cannot be parsed.</param>
        /// <returns>true when the value could be parsed; otherwise false</returns>
        public static bool TryParse(string s, out Money result) => 
            TryParse(s, NumberStyles.Number, NumberFormatInfo.CurrentInfo, out result);

        /// <summary>
        /// Parses the value and attempts to create a new <see cref="Money"/> from it.
        /// </summary>
        /// <param name="s">The value to parse. The value may omit the currencycode.</param>
        /// <param name="style">The numberstyle used to parse the numeric part of the value</param>
        /// <param name="provider">The formatprovider used to parse the numeric part of the value</param>
        /// <param name="result">The returned <see cref="Money"/> object. Will be in unspecified state when the value cannot be parsed.</param>
        /// <returns>true when the value could be parsed; otherwise false</returns>
        public static bool TryParse
        (
            string s,
            NumberStyles style,
            IFormatProvider provider,
            out Money result)
        {
            if (s is null)
            {
                result = Unspecified();
                return false;
            }

            var parts = s.Split(' ');
            if (parts.Length > 2)
            {
                result = Unspecified();
                return false;
            }

            var amountPart = parts[0];
            var currencyCodePart = parts.Length > 1 ? parts[1] : UNSPECIFIED_CURRENCY_CODE;
            if (currencyCodePart.Length != 3 || !decimal.TryParse(amountPart, style, provider, out var amount))
            {
                result = Unspecified();
                return false;
            }
            
            result = new Money(amount, currencyCodePart);
            return true;
        }
    }
}