﻿// Copyright 2011-2020 ReflexSystems

using System;
using System.ComponentModel;
using System.Globalization;

namespace StimulsoftValueType.ValueType
{
    /// <summary>
    /// A <see cref="TypeConverter"/> that can convert to and from <see cref="Money"/> object
    /// </summary>
    public sealed class MoneyConverter : TypeConverter
    {
        /// <inheritdoc />
        public override bool CanConvertFrom(
            ITypeDescriptorContext context,
            Type sourceType) =>
            sourceType == typeof(string) ||
            sourceType == typeof(double) ||
            sourceType == typeof(double?) ||
            sourceType == typeof(decimal) ||
            sourceType == typeof(decimal?) ||
            sourceType == typeof(object);

        /// <inheritdoc />
        public override object ConvertFrom(
            ITypeDescriptorContext context,
            CultureInfo culture,
            object value) =>
            value switch
            {
                decimal decimalValue => Money.Unspecified(decimalValue),
                double doubleValue => Money.Unspecified((decimal)doubleValue),
                string stringValue => Money.TryParse(stringValue, NumberStyles.Number, culture, out var money) ? money : (Money?)null,
                _ => null
            };

        /// <inheritdoc />
        public override bool CanConvertTo(
            ITypeDescriptorContext context,
            Type destinationType) =>
            destinationType == typeof(string) ||
            destinationType == typeof(double) ||
            destinationType == typeof(double?) ||
            destinationType == typeof(decimal) ||
            destinationType == typeof(decimal?) ||
            destinationType == typeof(object);
        
        /// <inheritdoc />
        public override object ConvertTo(
            ITypeDescriptorContext context,
            CultureInfo culture,
            object value,
            Type destinationType)
        {
            // When the value is null we can also return null
            // WPF won't supply decimal? or double? as a destination type so we don't know if we need to
            // return 0 or null
            if (value is null)
                return null;

            var money = (Money)value;
            if (destinationType == typeof(string))
                return money.ToString("C", culture);
            if (destinationType == typeof(double))
                return (double)money.Amount;
            if (destinationType == typeof(double?))
                return (double?)money.Amount;
            if (destinationType == typeof(decimal))
                return money.Amount;
            if (destinationType == typeof(decimal?))
                return (decimal?)money.Amount;
            return money;
        }
    }
}